/* $Header: /home/vikas/src/xtacacsd/RCS/taclast.c,v 1.5 1996/05/19 17:40:55 vikas Exp $ */
long calcredits();
/*
 * This is similar to the 'last' program on Unix systems but for processing
 * my xtacacsd utmp/wtmp files (xtacacsd runs on Cisco routers).
 * Also acts like a simple 'who' program to read the data in the tacacs UTMP
 * file.
 *
 *	-Vikas Aggarwal, vikas@navya.com, June 1995
 *
 * Derived from:
 *	$ NetBSD: last.c,v 1.6 1994/12/24 16:49:02 cgd Exp
 */


#ifndef lint
/* static char sccsid[] = "@(#)last.c	8.2 (Berkeley) 4/2/94"; */
/* static char rcsid[] = "$NetBSD: last.c,v 1.6 1994/12/24 16:49:02 cgd Exp "; */
static char rcsid[] = "$RCSfile: taclast.c,v $ $Revision: 1.5 $ $Date: 1996/05/19 17:40:55 $" ;
#endif /* not lint */

#include <sys/types.h>
#include <sys/param.h>
#include <sys/socket.h>		/* for MAXHOSTNAMELEN on SCO ? */
#include <sys/stat.h>

#include <fcntl.h>
#include <sys/file.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
/* #include <tzfile.h>			/* for SECSPERDAY, etc. */
#include <unistd.h>

#include "tacutmp.h"			/* our very own utmp.h file */

#ifndef	SECSPERDAY
#define	SECSPERDAY	((long) (60 * 60 * 24))
#endif
#ifndef DAYSPERWEEK
#define DAYSPERWEEK	7
#endif

#ifndef MAXHOSTNAMELEN
#define MAXHOSTNAMELEN 128
#endif

#if defined(SYSV) || defined(SVR4) || defined(__svr4__)
#  define index strchr
#  define bcopy(a, b, c) memmove(b, a, c)
#  define bzero(a, b)    memset(a, 0, b)
#endif

#define	NO	0				/* false/no */
#define	YES	1				/* true/yes */

static struct tacutmp	buf[1024];		/* utmp read buffer */

typedef struct arg {
	char	*name;				/* argument */
	time_t	totsecs;			/* total seconds */
	time_t	totcredits;			/* total timecredits */
#define	HOST_TYPE	-2
#define	TTY_TYPE	-3
#define	USER_TYPE	-4
	struct arg	*next;			/* linked list pointer */
} ARG;
ARG	*HOSTarglist, *USERarglist;		/* head of linked list */

typedef struct ttytab {
	time_t	logout;				/* log out time */
	int	waslogin;			/* boolean for prev entry */
	char	tty[UT_LINESIZE + 1];		/* terminal name */
	char	host[UT_HOSTSIZE + 1];		/* host name */
	char	name[UT_NAMESIZE + 1];		/* user on this TTY */
	struct ttytab	*next;			/* linked list pointer */
} TTY;
TTY	*ttylist;				/* head of linked list */

#ifndef WTMP
#define WTMP	"/var/adm/wtmp"
#define UTMP	"/etc/utmp"
#endif

static int	whomode;			/* do a 'who', not 'last' */
static int	ignorehost;			/* dont compare hostnames */
static int	dototals;			/* want totaltimes */
static int	noargs;				/* no cmd line args */
static time_t	currentout;			/* current logout value */
static long	maxrec;				/* records to display */
static char	*file;				/* wtmp file */
static char	*weekdays[] = {"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", 
				  "Sat", "Sun"};
void	 addarg (/* int, char * */);
TTY	*addtty (/* char * */);
void	 hostconv (/* char * */);
char	*ttyconv (/* char * */);
int	 want (/* struct tacutmp *, int */);
void	 wtmp (/* void */);

char	*prognm;

main(argc, argv)
	int argc;
	char *argv[];
{
	extern int optind;
	extern char *optarg;
	int ch;
	char *p;

	prognm = argv[0];	/* save program name */
	if (!strcmp(prognm, "tacwho"))		/* if called as tacwho */
	  whomode = 1;

	maxrec = -1;
	while ((ch = getopt(argc, argv, "0123456789f:h:itw")) != EOF)
		switch (ch) {
		case '0': case '1': case '2': case '3': case '4':
		case '5': case '6': case '7': case '8': case '9':
			/*
			 * kludge: last was originally designed to take
			 * a number after a dash.
			 */
			if (maxrec == -1) {
				p = argv[optind - 1];
				if (p[0] == '-' && p[1] == ch && !p[2])
					maxrec = atol(++p);
				else
					maxrec = atol(argv[optind] + 1);
				if (!maxrec)
					exit(0);
			}
			break;
		case 'f':
			file = optarg;
			break;
		case 'h':
/*			hostconv(optarg);	/* not used */
			addarg(HOST_TYPE, optarg);
			break;
		case 'i':
			ignorehost++;		/* if separate wtmp files */
			break;
		case 't':
		        dototals++;	/* if 2, then only print totals */
			break;
		case 'w':
			whomode++;
			break;
		case '?':
		default:
			(void)fprintf(stderr,
	"usage: %s [-# (max num records)] [-f file] [-h hostname] [-i (for ignore host)]", prognm);
			(void)fprintf(stderr,
	" [-w (for WHO mode)] [-t (totaltimes) [-t]] [user ...]\n");
			exit(1);
		}

	if (argc) {
#ifndef POSIX
		setlinebuf(stdout);	/* */
#else
		setbuf(stdout, NULL);	/* */
#endif
		for (argv += optind; *argv; ++argv) {
			addarg(USER_TYPE, *argv);
		}
	}

	if (!HOSTarglist && !USERarglist)
	  noargs = 1;			/* no cmd line arguments */

	if (whomode)
	  doutmp() ;
	else
	  wtmp();

	exit(0);
}

/*
 * wtmp --
 *	read through the wtmp file
 */
void
wtmp()
{
    struct tacutmp	*bp;		/* current structure */
    TTY	*T;				/* tty list entry */
    struct stat	stb;			/* stat of file for size */
    time_t	bl, delta;		/* time difference */
    int	bytes, wfd;
    char	*ct, *crmsg;

     if (file == (char *)NULL || *file == '\0')
       file = WTMP ;

    if ((wfd = open(file, O_RDONLY, 0)) < 0 || fstat(wfd, &stb) == -1) {
	fprintf(stderr, "%s: %s: ", prognm, file);
	perror ((char *)NULL);
	exit (1);
    }

    if ( (stb.st_size % sizeof buf[0]) != 0) {
	fprintf(stderr,
		"ERROR: WTMP record sizes not compatible with compiled tacutmp structure.\n");
	fprintf(stderr,
		"This is controlled using the XTACUTMP compile time define.\n");
#ifdef XTACUTMP
	fprintf(stderr, "Recompile program without XTACUTMP.\n");
#else
	fprintf(stderr, "Recompile program with XTACUTMP.\n");
#endif
	exit(1);
    }

    bl = (stb.st_size + sizeof(buf) - 1) / sizeof(buf);	/* total # of reads */

    (void)time(&buf[0].ut_time);

    while (--bl >= 0) {
	if (lseek(wfd, (off_t)(bl * sizeof(buf)), SEEK_SET) == -1 ||
	    (bytes = read(wfd, buf, sizeof(buf))) == -1) {
	    fprintf(stderr, "%s: %s: ", prognm, file);
	    perror((char *)NULL);
	    exit (1);
	}
	/* work backwards in the file */
	for (bp = &buf[bytes / sizeof(buf[0]) - 1]; bp >= buf; --bp) {
	    /*
	     * if the terminal line is '~', the machine stopped.
	     * see utmp(5) for more info.
	     */
	    if (bp->ut_line[0] == '~' && !bp->ut_line[1]) {
		/* everybody just logged out */
		for (T = ttylist; T; T = T->next) {
		    if (ignorehost || strncasecmp(T->host, bp->ut_host, UT_HOSTSIZE) == 0 )
		      T->logout = -bp->ut_time;	/* indicate stopped entry */
		}
		currentout = -bp->ut_time;
		crmsg = strncmp(bp->ut_name, "shutdown",
				UT_NAMESIZE) ? "crash" : "shutdown";
		if (want(bp, NO)) {
		    ct = ctime(&bp->ut_time);
		    printf("%-*.*s  %-*.*s %-*.*s %10.10s %5.5s \n",
			   UT_NAMESIZE, UT_NAMESIZE, bp->ut_name,
			   UT_LINESIZE, UT_LINESIZE, bp->ut_line,
			   UT_HOSTSIZE, UT_HOSTSIZE, bp->ut_host,
			   ct, ct + 11);
		    if (maxrec != -1 && !--maxrec)
		      return;
		}
		continue;
	    }	/* end if machine stopped */

	    /*
	     * if the line is '{' or '|', date got set; see
	     * utmp(5) for more info.
	     */
	    if ((bp->ut_line[0] == '{' || bp->ut_line[0] == '|')
		&& !bp->ut_line[1]) {
		if (want(bp, NO)) {
		    ct = ctime(&bp->ut_time);
		    printf("%-*.*s  %-*.*s %-*.*s %10.10s %5.5s \n",
			   UT_NAMESIZE, UT_NAMESIZE, bp->ut_name,
			   UT_LINESIZE, UT_LINESIZE, bp->ut_line,
			   UT_HOSTSIZE, UT_HOSTSIZE, bp->ut_host,
			   ct, ct + 11);
		    if (maxrec && !--maxrec)
		      return;
		}
		continue;
	    }	/* end if date got set */

	    /* find associated tty in TTY list */
	    for (T = ttylist;; T = T->next) {
		if (!T) {
		    /* add new one */
		    T = addtty(bp->ut_line, bp->ut_host, bp->ut_name);
		    break;
		}
		if (!strncmp(T->tty, bp->ut_line, UT_LINESIZE))
		  if (ignorehost || strncasecmp(T->host, bp->ut_host, UT_HOSTSIZE) == 0 )
		    break;
	    }

	    /* if this is a login entry */
	    if (bp->ut_name[0] && bp->ut_name[0] != UT_LOGOUT_CHAR && want(bp, YES))
	    {
		ct = ctime(&bp->ut_time);
		if (dototals < 2)
		  printf("%-*.*s  %-*.*s %-*.*s %10.10s %5.5s ",
			 UT_NAMESIZE, UT_NAMESIZE, bp->ut_name,
			 UT_LINESIZE, UT_LINESIZE, bp->ut_line,
			 UT_HOSTSIZE, UT_HOSTSIZE, bp->ut_host,
			 ct, ct + 11);
		if (!T->logout) {
		    if (dototals < 2) printf(" online");
		}
		else if (T->logout < 0) {	/* stored as -ve on reboot */
		    T->logout = -T->logout;
		    if (dototals < 2) printf("- %s", crmsg);	/* shutdown or crash */
		}
		else
		  if (dototals < 2) printf("- %5.5s", ctime(&T->logout)+11);

		delta = (T->logout ? T->logout:time(NULL)) - bp->ut_time;

if (!T->logout) printf("Nologout\n");
else {
printf("Delta is:%ld %ld %ld\n",delta,T->logout,bp->ut_time);
printf("%ld %s",T->logout,ctime(&T->logout));
printf("%ld %s",bp->ut_time,ctime(&bp->ut_time));
}
		if (delta < 0) {	/* should not happen */
		    if (dototals < 2) printf(" (negative)");
		    delta = 0;
		}
		else if (delta < SECSPERDAY) {
		    if (dototals < 2) printf("  (%5.5s)", asctime(gmtime(&delta))+11);
		}
		else
		  if (dototals < 2) printf(" (%ld+%5.5s)", delta / SECSPERDAY,
			 asctime(gmtime(&delta))+11);
		if (dototals < 2) printf(" %ld", delta); /* actual num secs */
		/* Check for record correctness- prev was also a login ? */
		if (T->waslogin || (bp->ut_name[0] == UT_LOGOUT_CHAR && strncmp ((bp->ut_name)+1, (T->name)+1, sizeof(bp->ut_name) - 1) ) )
		  if (dototals < 2) printf (" inaccurate");	/* missing entries */
		if (dototals < 2) printf("\n");

		if (dototals)		/* total up the times */ /*Add credits here*/
		  addsecs(bp, delta,calcredits((T->logout?T->logout:time(NULL)),bp->ut_time));

		if (maxrec != -1 && !--maxrec)
		  return;
	    }	/* end if login entry */

	    /* 	We always store this bp->ut_time. If was login entry, then
	     *	serves as the logout in case the logout entry is lost for the
	     *  previous person
	     */
	    T->logout = bp->ut_time;
	    if (bp->ut_name[0] && bp->ut_name[0] != UT_LOGOUT_CHAR)
	      T->waslogin = 1 ;
	    else
	      T->waslogin = 0;

	}	/* end for() work backwards in file */
    }	/* end while(--bl) */

    if (dototals)
      pr_totaltimes();

    ct = ctime(&buf[0].ut_time);
    printf("\nwtmp begins %10.10s %5.5s \n", ct, ct + 11);

}	/* end: main() */


/*
 * want --
 *	see if want this entry
 */
int
want(bp, check)
     struct tacutmp *bp;
     int check;		/* check if username is 'ftp' or 'uucp' */
{
    ARG *step;

    if (check)
      /*
       * when uucp and ftp log in over a network, the entry in
       * the utmp file is the name plus their process id.  See
       * etc/ftpd.c and usr.bin/uucp/uucpd.c for more information.
       */
      if (!strncmp(bp->ut_line, "ftp", sizeof("ftp") - 1))
	bp->ut_line[3] = '\0';
      else if (!strncmp(bp->ut_line, "uucp", sizeof("uucp") - 1))
	bp->ut_line[4] = '\0';

    if (noargs)
      return (YES);
    
    for (step = HOSTarglist; step; step = step->next)
      if (!strncasecmp(step->name, bp->ut_host, UT_HOSTSIZE))
	return (YES);

    for (step = USERarglist; step; step = step->next)
      if (!strncmp(step->name, bp->ut_name, UT_NAMESIZE))
	return (YES);
    
    return (NO);

}

/*
 * addarg --
 *	add an entry to a linked list of arguments
 */
void
addarg(type, arg)
	int type;
	char *arg;
{
	ARG *cur, **whichlist;

	if (!(cur = (ARG *)malloc((u_int)sizeof(ARG)))) {
		fprintf(stderr, "%s: malloc failure.\n", prognm);
		exit (1);
	    }
	switch (type) {
	 case HOST_TYPE:
	    whichlist = &HOSTarglist;	
	    break;
	 case USER_TYPE:
	    whichlist = &USERarglist;
	    break;
	 default:
	    fprintf(stderr, "addarg: unknown type %d\n", type);
	}
	cur->next = *whichlist;
	cur->totsecs = 0;
	cur->name = arg;
	*whichlist = cur;
}

/*
 * addtty --
 *	add an entry to a linked list of ttys
 */
TTY *
addtty(ttyname, hostname, username)
	char *ttyname, *hostname, *username;
{
	TTY *cur;

	if (!(cur = (TTY *)malloc((u_int)sizeof(TTY)))) {
		fprintf(stderr, "%s: malloc failure.\n", prognm);
		exit (1);
	    }
	cur->next = ttylist;
	cur->logout = currentout;
	cur->waslogin = 0;	/* initialization- logout */
	bcopy(ttyname, cur->tty, UT_LINESIZE);
	bcopy(hostname, cur->host, UT_HOSTSIZE);
	bcopy(username, cur->name, UT_NAMESIZE);
	return (ttylist = cur);
}

/*
 * Add the delta seconds to the total number of seconds. Add to the arglist
 * if this is a new host or name and nothing was supplied on the command line.
 */
addsecs(bp, delta,credits)
     struct tacutmp *bp;
     time_t delta;
     long credits;
{
    ARG *step;

    for (step = HOSTarglist; step; step = step->next)
      if (!strncasecmp(step->name, bp->ut_host, UT_HOSTSIZE)) {
	  step->totsecs += (long)delta;
	  break ;
      }
    if (!step && noargs) {		/* new host, add to linked list */
	char *s = (char *)malloc(UT_HOSTSIZE + 1);
	strncpy (s, bp->ut_host, UT_HOSTSIZE);
	addarg(HOST_TYPE, s);
	HOSTarglist->totsecs += (long)delta;
    }

    for (step = USERarglist; step; step = step->next)
      if (!strncmp(step->name, bp->ut_name, UT_NAMESIZE)) {
	  step->totsecs += (long)delta;
	  step->totcredits += (long)credits;
	  break ;
      }
    if (!step && noargs) {
	char *s = (char *)malloc(UT_NAMESIZE + 1);
	strncpy (s, bp->ut_name, UT_NAMESIZE);
      	addarg(USER_TYPE, s);
	USERarglist->totsecs += (long)delta;
	USERarglist->totcredits += (long)credits;
    }

}	/* end: add_secs() */

/*
 * print the total time for each user supplied argument
 */

pr_totaltimes()
{
    ARG *step;
    
    printf("\nTOTAL TIMES (d+hh:mm)\n===========\n");
    
    for (step = USERarglist; step; step = step->next)
    {
	time_t tsecs = step->totsecs ;
	time_t tcreds = step->totcredits ;

	printf("USER %s", step->name);
	if (tsecs < SECSPERDAY)
	  printf("  (%5.5s)", asctime(gmtime(&tsecs))+11);
	else
	  printf(" (%ld+%5.5s)",
		 tsecs / SECSPERDAY, asctime(gmtime(&tsecs))+11);
	printf(" %ld\n", tsecs);
        printf(" Credits: %ld\n", tcreds);
    }	/* for() */
    for (step = HOSTarglist; step; step = step->next)
    {
	time_t tsecs = step->totsecs ;

	printf("HOST %s", step->name);
	if (tsecs < SECSPERDAY)
	  printf("  (%5.5s)", asctime(gmtime(&tsecs))+11);
	else
	  printf(" (%ld+%5.5s)",
		 tsecs / SECSPERDAY, asctime(gmtime(&tsecs))+11);
	printf(" %ld\n", tsecs);
    }	/* for() */

}	/* pr_totaltimes() */


/*
 * hostconv --
 *	convert the hostname to search pattern; if the supplied host name
 *	has a domain attached that is the same as the current domain, rip
 *	off the domain suffix since that's what login(1) does.
 */
void
hostconv(arg)
	char *arg;
{
	static int first = 1;
	static char *hostdot, name[MAXHOSTNAMELEN];
	char *argdot;

	if (!(argdot = strchr(arg, '.')))
		return;
	if (first) {
		first = 0;
		if (gethostname(name, sizeof(name))) {
			perror("last: gethostname");
			exit (1);
		    }
		hostdot = strchr(name, '.');
	}
	if (hostdot && !strcasecmp(hostdot, argdot))
		*argdot = '\0';
}

/*
 * Following are 'who' routines for printing out the data in the
 * UTMP file to show users currently logged in.
 */

doutmp()
{
    time_t now, delta;
    struct tacutmp u ;
    FILE   *ut;
    u_long l;
    int    nusers = 0, tusers = 0;	/* wanted users, total users */

    now = time(NULL);

    if (file == (char *)NULL || *file == '\0')
      file = UTMP ;
    if ((ut = fopen(file, "r")) == NULL) {
	fprintf(stderr, "%s: %s: ", prognm, file);
	perror ((char *)NULL);
	exit (1);
    }

#define HEADER	"USER     TTY      FROM              LOGIN@  ONLINE(d+h:m)\n"
    (void)printf(HEADER);

    while (fread(&u, sizeof(u), 1, ut)) {
	if (u.ut_name[0] == '\0' || u.ut_name[0] == UT_LOGOUT_CHAR)
	  continue;
	++tusers;
	if (want(&u, NO))  {
	    (void)printf("%-*.*s %-*.*s %-*.*s ",
			 UT_NAMESIZE, UT_NAMESIZE, u.ut_name,
			 UT_LINESIZE, UT_LINESIZE, u.ut_line,
			 UT_HOSTSIZE, UT_HOSTSIZE, *(u.ut_host) ? u.ut_host : "-");
	    delta = now - u.ut_time;
	    pr_attime(&delta, &(u.ut_time), &now); /* */
	    if (delta < SECSPERDAY)
	      printf("  (%5.5s)",
		     asctime(gmtime(&delta))+11);
	    else
	      printf(" (%ld+%5.5s)",
		     delta / SECSPERDAY,
		     asctime(gmtime(&delta))+11);
	    printf("\n");
	    ++nusers;
	    continue;
	}

	(void)fclose(ut);
	(void)printf("\nTotal users: %d (%d)\n", nusers, tusers);

    }	/* end: for */

}
/*
 * pr_attime --
 *	Print the time that the user logged in
 */

pr_attime(delta, started, now)
	time_t *delta, *started, *now;
{
	static char buf[256];
	struct tm *tp;

	tp = localtime(started);

	if (*delta > (SECSPERDAY * DAYSPERWEEK))	/* more than a week */
	  printf("%2.2d-%2.2d-%2.2d",
		 tp->tm_mday, tp->tm_mon + 1, tp->tm_year);

	else if (*now / SECSPERDAY != *started / SECSPERDAY) {	/* not today */
	    printf("%s-%2.2d:%2.2d",
		  weekdays[tp->tm_wday], tp->tm_hour, tp->tm_min);
	}

	else {	/* today */
	    printf("%2.2d:%2.2d", tp->tm_hour, tp->tm_min);
	}

}	/* end pr_attime() */


long calcredits(long start,long end)
{
long credits=0;

    credits=start-end;

return credits;
}
